from datetime import datetime, timedelta
import json
import os
import time
from typing import Any, Dict, Optional
from ksxt.api.koreainvest import ImplicitAPI
from ksxt.async_.base.async_exchange import AsyncExchange
from ksxt.market.manager import MarketManager
from ksxt.parser.koreainvest import KoreaInvestParser


class KoreaInvest(AsyncExchange, ImplicitAPI):
    def __init__(self, config: Dict = None) -> None:
        super().__init__(config, "koreainvest.toml")
        self.parser = KoreaInvestParser()

    def safe_symbol(self, base_market: str, security: str) -> str:
        return f"{security}"

    def is_activate(self, path, security_type) -> bool:
        mode = "dev" if self.is_dev == True else "app"

        tr_id = self.apis[self.type][security_type][path][mode]["tr_id"]

        if not bool(tr_id):
            return False

        return super().is_activate(path=path, security_type=security_type)

    @AsyncExchange.check_token
    def sign(
        self,
        path,
        security_type,
        method_type,
        api_type: Any = "public",
        headers: Optional[Any] = None,
        body: Optional[Any] = None,
        params: Optional[Any] = None,
        config={},
    ):
        mode = "dev" if self.is_dev == True else "app"

        host_url = self.apis[self.type][mode]["hostname"]
        destination = self.apis[self.type][security_type][path]["url"]
        version = self.apis[self.type]["version"]
        params["version"] = version
        destination = self.implode_params(destination, params)

        url = host_url + "/" + destination

        tr_id = self.apis[self.type][security_type][path][mode]["tr_id"]
        authorization_token = f"Bearer {self.token}"
        custtype = ""
        if "custtype" in self.apis[self.type][security_type][path]:
            custtype = self.apis[self.type][security_type][path]["custtype"]

        if headers is None:
            headers = {}
            headers.update(
                {
                    "content-type": "application/json",
                    "authorization": authorization_token,
                    "appkey": self.open_key,
                    "appsecret": self.secret_key,
                    "tr_id": tr_id,
                    "custtype": custtype,
                }
            )

        if method_type.upper() == "POST":
            body = json.dumps(params)
            params = {}

        return {"url": url, "method": method_type, "headers": headers, "body": body, "params": params}

    # region _____
    def create_token(self):
        import logging

        mode = "dev" if self.is_dev == True else "app"
        host_url = self.apis[self.type][mode]["hostname"]
        destination = self.apis[self.type]["token"]["url"]
        url = host_url + "/" + destination

        request_headers = self.prepare_request_headers()

        body = {"grant_type": "client_credentials", "appkey": self.open_key, "appsecret": self.secret_key}

        body = json.dumps(body, separators=(",", ":"))

        try:
            res = self.fetch(url=url, method="POST", headers=request_headers, body=body)
        except Exception as ex:
            logging.error(ex)
            return

        if "access_token" in res.keys():
            token = res["access_token"]
            token_expired = res["access_token_token_expired"]
            self.save_token(self.open_key, token, token_expired)
        else:
            logging.error(res["error_description"])

    def get_response(self, result, request_params: dict):
        if self.safe_string(result, "rt_cd") == "0":
            response = self.get_success_response(result)
        else:
            response = self.get_error_response(result)

        response = self.update_header(response, request_params)
        return response

    def get_success_response(self, result):
        return super().get_success_response(
            # 성공 실패 여부
            response_code="0",
            # 응답 코드
            msg_code=self.safe_string(result, "msg_cd"),
            # 응답 메세지
            msg=self.safe_string(result, "msg1"),
            # 원본 데이터
            org_data=result,
        )

    def get_error_response(self, result):
        return super().get_error_response(
            error_code=self.safe_string(result, "msg_cd"),
            error_message=self.safe_string(result, "msg1"),
            org_data=result,
        )

    def update_header(self, response: dict, request_params: dict):
        response = self.update_success_response_header(
            response, {"request_time": time.time(), "request_params": request_params}
        )
        return response

    # endregion ____

    # region public feeder
    @AsyncExchange.check_token
    async def fetch_markets(self, market_name: str):
        db_path = os.path.join(os.getcwd(), f".ksxt-cache_market.{datetime.now().strftime('%Y%m%d')}.db")
        manager = MarketManager(db_path=db_path)
        manager._init()
        if market_name.lower() == "kospi":
            result = manager.kospi.all()
            base_market = "KRW"
        elif market_name.lower() == "kosdaq":
            result = manager.kosdaq.all()
            base_market = "KRW"
        elif market_name.lower() == "nyse":
            result = manager.nyse.all()
            base_market = "USD"
        elif market_name.lower() == "nasdaq":
            result = manager.nasdaq.all()
            base_market = "USD"
        elif market_name.lower() == "amex":
            result = manager.amex.all()
            base_market = "USD"
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{market_name} market is not yet supported."
            )

        params = {}
        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_markets(response=response, base_market=base_market)

    async def fetch_trade_fee(self, symbol: Optional[str] = "", base_market: str = "KRW"):
        params = {"base_market": base_market}

        result = {"rt_cd": "0", "output": {}}
        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_trade_fee(response, base_market)

    async def fetch_ticker(self, symbol: str, base_market: str = "KRW"):
        if base_market == "KRW":
            params = {"FID_COND_MRKT_DIV_CODE": "J", "FID_INPUT_ISCD": symbol}
            result = await self.private_get_fetch_ticker_price_krw(self.extend(params))
        elif base_market == "USD":
            market_code = await self.get_market_code_in_feeder(symbol=symbol, base_market=base_market)
            params = {"AUTH": "", "EXCD": market_code, "SYMB": symbol}
            result = await self.fetchTickerForUS(self.extend(params))
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_ticker(response, base_market)

    async def fetch_historical_data(
        self,
        symbol: str,
        time_frame: str,
        start: Optional[str] = None,
        end: Optional[str] = None,
        base_market: str = "KRW",
    ):
        limit = 100

        if end is None:
            end = KoreaInvest.now(base_market=base_market)

        if start is None:

            if time_frame.endswith("m"):
                start = end - timedelta(minutes=30)

                if base_market == "KRW":
                    params = {
                        "FID_ETC_CLS_CODE": "",
                        "FID_COND_MRKT_DIV_CODE": "J",  # "U"
                        "FID_INPUT_ISCD": symbol,
                        "FID_INPUT_HOUR_1": start.strftime("%H%M%S"),  # 업종은 조회 간격만 입력 가능
                        "FID_PW_DATA_INCU_YN": "N",
                    }

                    result = await self.private_get_fetch_security_ohlcv_minute_krw(self.extend(params))

            elif time_frame.endswith("D"):
                start = end - timedelta(days=limit)
            elif time_frame.endswith("W"):
                start = end - timedelta(weeks=limit)
            elif time_frame.endswith("M"):
                start = end - timedelta(days=limit * 30)
            elif time_frame.endswith("Y"):
                start = end - timedelta(days=limit * 365)
            else:
                start = end

        if base_market == "KRW":
            params = {
                "FID_COND_MRKT_DIV_CODE": "J",
                "FID_INPUT_ISCD": symbol,
                "FID_INPUT_DATE_1": start.strftime("%Y%m%d"),
                "FID_INPUT_DATE_2": end.strftime("%Y%m%d"),
                "FID_PERIOD_DIV_CODE": time_frame,
                "FID_ORG_ADJ_PRC": "1",
            }

            result = await self.private_get_fetch_security_ohlcv_krw(self.extend(params))
            # index는 구분해서 날려야함
            # response = self.private_get_fetch_index_ohlcv_krw(self.extend(params))
        elif base_market == "USD":
            if time_frame == "D":
                gubn = "0"
            elif time_frame == "W":
                gubn = "1"
            elif time_frame == "M":
                gubn = "2"
            else:
                return self.get_error_response(
                    error_code="time frame error", error_message=f"{time_frame} time-frame is not supported."
                )

            market_code = await self.get_market_code_in_feeder(symbol=symbol, base_market=base_market)

            params = {
                "AUTH": "",
                "EXCD": market_code,
                "SYMB": symbol,
                "GUBN": gubn,
                "BYMD": end.strftime("%Y%m%d"),
                "MODP": "1",
                "KEYB": "",
            }

            result = await self.fetchOHLCVforUS(self.extend(params))
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_historical_data(response, base_market)

    async def fetch_is_holiday(self, dt: datetime, base_market: str = "KRW"):
        if base_market == "KRW":
            params = {"BASS_DT": dt.strftime("%Y%m%d"), "CTX_AREA_NK": "", "CTX_AREA_FK": ""}
            result = await self.private_get_fetch_calendar_holiday_krw(self.extend(params))

            response = self.get_response(result, params)
            if response["response"]["success"] != "0":
                return response

            return self.parser.parse_is_holiday(response, base_market)
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

    # endregion public feeder

    # region private feeder
    # @RestExchange.check_token
    async def fetch_balance(self, acc_num: str, base_market: str = "KRW"):
        if base_market == "KRW":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "AFHR_FLPR_YN": "N",
                "OFL_YN": "",
                "INQR_DVSN": "01",
                "UNPR_DVSN": "01",
                "FUND_STTL_ICLD_YN": "N",
                "FNCG_AMT_AUTO_RDPT_YN": "N",
                "PRCS_DVSN": "01",
                "CTX_AREA_FK100": "",
                "CTX_AREA_NK100": "",
            }

            result = await self.private_get_fetch_balance_krw(self.extend(params))
        elif base_market == "USD":
            market_code = await self.get_market_code_in_feeder(symbol="ALL", base_market=base_market)
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "OVRS_EXCG_CD": market_code,
                "TR_CRCY_CD": "USD",
                "CTX_AREA_FK200": "",
                "CTX_AREA_NK200": "",
            }

            result = await self.private_get_fetch_balance_usd(self.extend(params))
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_balance(response, base_market)

    async def fetch_cash(self, acc_num: str, base_market: str = "KRW"):
        if base_market == "KRW":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "AFHR_FLPR_YN": "N",
                "OFL_YN": "",
                "INQR_DVSN": "01",
                "UNPR_DVSN": "01",
                "FUND_STTL_ICLD_YN": "N",
                "FNCG_AMT_AUTO_RDPT_YN": "N",
                "PRCS_DVSN": "01",
                "CTX_AREA_FK100": "",
                "CTX_AREA_NK100": "",
            }

            result = await self.private_get_fetch_cash_krw(self.extend(params))
        elif base_market == "USD":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                # 01: 원화, 02: 외화
                "WCRC_FRCR_DVSN_CD": "02",
                "NATN_CD": "840",
                "TR_MKET_CD": "00",
                "INQR_DVSN_CD": "00",
            }

            result = await self.fetchCashForUS(self.extend(params))
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_cash(response, base_market)

    async def fetch_orderbook(self, symbol: str, base_market: str = "KRW"):
        if base_market == "KRW":
            params = {"FID_COND_MRKT_DIV_CODE": "J", "FID_INPUT_ISCD": symbol}

            result = await self.private_get_fetch_orderbook_krw(self.extend(params))
        elif base_market == "USD":
            params = {"FID_COND_MRKT_DIV_CODE": "J", "FID_INPUT_ISCD": symbol}

            result = await self.fetchCashForUS(self.extend(params))
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_orderbook(response, base_market)

    async def fetch_pnl(self, acc_num: str, include_cost: bool = True, base_market: str = "KRW"):
        if base_market == "KRW":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "AFHR_FLPR_YN": "N",
                "OFL_YN": "",
                "INQR_DVSN": "00",
                "UNPR_DVSN": "01",
                "FUND_STTL_ICLD_YN": "N",
                "FNCG_AMT_AUTO_RDPT_YN": "N",
                "PRCS_DVSN": "01",
                "COST_ICLD_YN": "Y" if include_cost else "N",
                "CTX_AREA_FK100": "",
                "CTX_AREA_NK100": "",
            }

            result = await self.private_get_fetch_pnl_krw(self.extend(params))
        elif base_market == "USD":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                # 01: 원화, 02: 외화
                "WCRC_FRCR_DVSN_CD": "02",
                "NATN_CD": "840",
                "TR_MKET_CD": "00",
                "INQR_DVSN_CD": "00",
            }

            result = await self.fetchCashForUS(self.extend(params))
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_pnl(response, base_market)

    async def fetch_screener_list(self, user_id, base_market: str = "KRW"):
        params = {"USER_ID": user_id}

        if base_market == "KRW":
            result = await self.private_get_fetch_screener_list_krw(self.extend(params))
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_screener_list(response, base_market)

    async def fetch_screener(self, user_id: str, screen_id: str, base_market: str = "KRW"):
        if base_market == "KRW":
            params = {"USER_ID": user_id, "SEQ": screen_id}

            result = await self.private_get_fetch_screener_krw(self.extend(params))
        elif base_market == "USD":
            market_code = self.get_market_code_in_feeder(symbol="ALL", base_market=base_market)
            params = {"AUTH": "", "EXCD": market_code, "CO_YN_PRICECUR": 1}
            result = await self.fetchScreenerForUS(self.extend(params))

        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_screener(response, base_market)

    async def fetch_security(self, symbol: str, base_market: str = "KRW"):
        if base_market == "KRW":
            params = {"PRDT_TYPE_CD": "300", "PDNO": symbol}  # 주식

            result = await self.private_get_fetch_security_info_krw(self.extend(params))
        elif base_market == "USD":
            params = {"PRDT_TYPE_CD": "J", "PDNO": symbol}

            result = await self.fetchCashForUS(self.extend(params))
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_security(response, base_market)

    # endregion private feeder

    # region broker
    async def create_order(
        self,
        acc_num: str,
        symbol: str,
        ticket_type: str,
        price: float,
        qty: float,
        otype: str,
        base_market: str = "KRW",
    ):
        if base_market == "KRW":
            if otype.upper() == "limit".upper():
                order_dvsn = "00"
            elif otype.upper() == "market".upper():
                order_dvsn = "01"

            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "PDNO": symbol,
                "ORD_DVSN": order_dvsn,
                "ORD_QTY": str(qty),  # string type 으로 설정
                "ORD_UNPR": str(price),  # string type 으로 설정
            }

            if ticket_type == "EntryLong":
                result = await self.private_post_send_order_entry_krw(self.extend(params))
            elif ticket_type == "ExitLong":
                result = await self.private_post_send_order_exit_krw(self.extend(params))
            else:
                return
        elif base_market == "USD":
            if otype.upper() == "limit".upper():
                order_dvsn = "00"
            elif otype.upper() == "market".upper():
                # 미국장은 시장가를 세부적으로 구분하여 지원함. -> 시장가 거래를 우선 지원하지 않는다.
                # https://apiportal.koreainvestment.com/apiservice/apiservice-overseas-stock#L_e4a7e5fd-eed5-4a85-93f0-f46b804dae5f
                return self.get_error_response(
                    error_code="market_error", error_message=f"{base_market} market is not yet supported."
                )

            if ticket_type == "entry_long":
                sell_type = ""
            elif ticket_type == "exit_long":
                sell_type = "00"
            else:
                return

            market_code = await self.get_market_code_in_broker(symbol=symbol, base_market=base_market)
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "OVRS_EXCG_CD": market_code,
                "PDNO": symbol,
                "ORD_DVSN": order_dvsn,
                "ORD_QTY": str(qty),  # string type 으로 설정
                "SLL_TYPE": sell_type,
                "OVRS_ORD_UNPR": str(price),  # string type 으로 설정
                "ORD_SVR_DVSN_CD": "0",
            }

            if ticket_type == "entry_long":
                result = await self.sendOrderEntryForUS(self.extend(params))
            elif ticket_type == "exit_long":
                result = await self.sendOrderExitForUS(self.extend(params))
            else:
                return
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_create_order(response, base_market)

    async def cancel_order(
        self,
        acc_num: str,
        order_id: str,
        symbol: Optional[str] = "",
        qty: float = 0,
        base_market: str = "KRW",
        **kwargs,
    ):
        if base_market == "KRW":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "KRX_FWDG_ORD_ORGNO": "",
                "ORGN_ODNO": str(order_id),
                "RVSE_CNCL_DVSN_CD": "02",
                "ORD_DVSN": "00",
                "ORD_QTY": str(qty),
                "ORD_UNPR": str(0),
                "QTY_ALL_ORD_YN": "N",
            }

            # 수량 미입력시 전량 취소
            if qty == 0:
                params["QTY_ALL_ORD_YN"] = "Y"

            result = await self.private_post_send_cancel_order_krw(self.extend(params))
        elif base_market == "USD":
            if qty == 0:
                return self.get_error_response(
                    error_code="qty_error", error_message=f"{base_market} cancel order need to set qty."
                )

            market_code = await self.get_market_code_in_broker(symbol=symbol, base_market=base_market)
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "OVRS_EXCG_CD": market_code,
                "PDNO": symbol,
                "ORGN_ODNO": str(order_id),
                "RVSE_CNCL_DVSN_CD": "02",
                "ORD_QTY": str(qty),
                "OVRS_ORD_UNPR": str(0),
            }

            result = await self.sendCancelOrderForUS(self.extend(params))
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_cancel_order(response, base_market)

    async def modify_order(
        self,
        acc_num: str,
        order_id: str,
        price: float,
        qty: float,
        symbol: Optional[str] = "",
        base_market: str = "KRW",
        **kwargs,
    ):
        if base_market == "KRW":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "KRX_FWDG_ORD_ORGNO": "",
                "ORGN_ODNO": str(order_id),
                "RVSE_CNCL_DVSN_CD": "01",
                "ORD_DVSN": "00",
                "ORD_QTY": str(qty),
                "ORD_UNPR": str(price),
                "QTY_ALL_ORD_YN": "N",
            }

            # 수량 미입력시 전량 수정
            if qty == 0:
                params["QTY_ALL_ORD_YN"] = "Y"

            result = await self.private_post_send_modify_order_krw(self.extend(params))
        elif base_market == "USD":
            market_code = await self.get_market_code_in_broker(symbol=symbol, base_market=base_market)
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "OVRS_EXCG_CD": market_code,
                "PDNO": symbol,
                "ORGN_ODNO": str(order_id),
                "RVSE_CNCL_DVSN_CD": "01",
                "ORD_QTY": str(qty),
                "OVRS_ORD_UNPR": str(price),
            }

            result = await self.sendModifyOrderForUS(self.extend(params))
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_modify_order(response, base_market)

    async def fetch_open_order(
        self,
        acc_num: str,
        symbol: Optional[str] = "",
        start: Optional[str] = None,
        end: Optional[str] = None,
        base_market: str = "KRW",
    ):
        if start is None:
            start = KoreaInvest.now(base_market=base_market).strftime("%Y%m%d")

        if end is None:
            end = KoreaInvest.now(base_market=base_market).strftime("%Y%m%d")

        if base_market == "KRW":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "INQR_STRT_DT": start,
                "INQR_END_DT": end,
                "SLL_BUY_DVSN_CD": "00",
                "INQR_DVSN": "00",
                "PDNO": symbol,
                "CCLD_DVSN": "02",
                "ORD_GNO_BRNO": "",
                "ODNO": "",
                "INQR_DVSN_3": "00",
                "INQR_DVSN_1": "",
                "CTX_AREA_FK100": "",
                "CTX_AREA_NK100": "",
            }

            result = await self.private_get_fetch_opened_order_krw(self.extend(params))
        elif base_market == "USD":
            market_code = await self.get_market_code_in_broker("ALL", base_market=base_market)
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "PDNO": symbol if symbol is not None else "%",
                "OVRS_EXCG_CD": market_code,
                "SORT_SQN": "DS",  # DS : 정순, AS : 역순
                "CTX_AREA_NK200": "",
                "CTX_AREA_FK200": "",
            }

            result = await self.fetchOpenedOrderForUS(self.extend(params))
        else:
            return self.get_error_response(
                error_code="market_error", error_message=f"{base_market} market is not yet supported."
            )

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_open_order_history(response, base_market)

    async def fetch_closed_order(
        self,
        acc_num: str,
        symbol: Optional[str] = "",
        start: Optional[str] = None,
        end: Optional[str] = None,
        base_market: str = "KRW",
    ):
        if start is None:
            start = KoreaInvest.now(base_market=base_market).strftime("%Y%m%d")

        if end is None:
            end = KoreaInvest.now(base_market=base_market).strftime("%Y%m%d")

        if base_market == "KRW":
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "INQR_STRT_DT": start,
                "INQR_END_DT": end,
                "SLL_BUY_DVSN_CD": "00",
                "INQR_DVSN": "00",
                "PDNO": symbol,
                "CCLD_DVSN": "01",
                "ORD_GNO_BRNO": "",
                "ODNO": "",
                "INQR_DVSN_3": "00",
                "INQR_DVSN_1": "",
                "CTX_AREA_FK100": "",
                "CTX_AREA_NK100": "",
            }

            result = await self.private_get_fetch_closed_order_krw(self.extend(params))
        elif base_market == "USD":
            market_code = await self.get_market_code_in_broker("ALL", base_market=base_market)
            params = {
                "CANO": acc_num[:8],
                "ACNT_PRDT_CD": acc_num[-2:],
                "PDNO": symbol if symbol is not None else "%",
                "ORD_STRT_DT": start,
                "ORD_END_DT": end,
                "SLL_BUY_DVSN": "00",
                "CCLD_NCCS_DVSN": "01" if not self.is_dev else "00",
                "OVRS_EXCG_CD": market_code,
                "SORT_SQN": "DS",  # DS : 정순, AS : 역순
                "ORD_DT": "",
                "ORD_GNO_BRNO": "",
                "ODNO": "",
                "CTX_AREA_NK200": "",
                "CTX_AREA_FK200": "",
            }

            result = await self.fetchClosedOrderForUS(self.extend(params))

        response = self.get_response(result, params)
        if response["response"]["success"] != "0":
            return response

        return self.parser.parse_closed_order_history(response, base_market)

    # endregion broker

    async def get_market_code_in_feeder(self, symbol: str, base_market: str = "KRW"):
        if base_market == "KRW":
            return ""
        elif base_market == "USD":
            if symbol.upper() == "ALL":
                return "NASD"

            response = await self.fetch_security(symbol=symbol, base_market=base_market)
            return response["exchange"]
        else:
            return ""

    async def get_market_code_in_broker(self, symbol: str, base_market: str = "KRW"):
        if base_market == "KRW":
            return ""
        elif base_market == "USD":
            if symbol.upper() == "ALL":
                return "NASD"

            response = await self.fetch_security(symbol=symbol, base_market=base_market)
            exname = response["exchange"]
            if exname == "NYS":
                return "NYSE"
            elif exname == "NAS":
                return "NASD"
            elif exname == "AMS":
                return "AMEX"
            else:
                return ""
        else:
            return ""



import asyncio
import collections
from time import time


class Throttler:
    def __init__(self, config, loop=None):
        self.loop = loop
        self.config = {
            "refillRate": 1.0,
            "delay": 0.001,
            "cost": 1.0,
            "tokens": 0,
            "maxCapacity": 2000,
            "capacity": 1.0,
        }

        self.config.update(config)
        self.queue = collections.deque()
        self.running = False

    async def looper(self):
        last_timestamp = time() * 1000
        while self.running:
            future, cost = self.queue[0]
            cost = self.config["cost"] if cost is None else cost
            if self.config["tokens"] >= 0:
                self.config["tokens"] -= cost

                if not future.done():
                    future.set_result(None)

                self.queue.popleft()

                # context switch
                await asyncio.sleep(0)

                if len(self.queue) == 0:
                    self.running = False
            else:
                await asyncio.sleep(self.config["delay"])
                now = time() * 1000
                elapsed = now - last_timestamp
                last_timestamp = now
                self.config["tokens"] = min(
                    self.config["tokens"] + elapsed * self.config["refillRate"], self.config["capacity"]
                )

    def __call__(self, cost=None):
        future = asyncio.Future()
        if len(self.queue) > self.config["maxCapacity"]:
            raise RuntimeError(
                "throttle queue is over maxCapacity ("
                + str(int(self.config["maxCapacity"]))
                + "), see https://github.com/ccxt/ccxt/issues/11645#issuecomment-1195695526"
            )

        self.queue.append((future, cost))
        if not self.running:
            self.running = True
            asyncio.ensure_future(self.looper(), loop=self.loop)

        return future



import hashlib
import json
import time
import uuid
from datetime import datetime
from typing import Any, Dict, List, Literal, Optional
from urllib.parse import urlencode

import jwt
import pytz

from ksxt.api.upbit import ImplicitAPI
from ksxt.parser.upbit import UpbitParser
from ksxt.base.rest_exchange import RestExchange
import ksxt.models


class Upbit(RestExchange, ImplicitAPI):
    def __init__(self, config: Dict = None) -> None:
        super().__init__(config, "upbit.toml")
        self.parser = UpbitParser()
        self.timezone = pytz.timezone("Asia/Seoul")

    def safe_symbol(self, base_market: str, security: str) -> str:
        # If security already contains a hyphen, assume it's correctly formatted
        if "-" in security:
            return security

        # Otherwise, assume security is just the symbol and format it with base_market
        return f"{base_market}-{security}"

    def safe_security(self, symbol: str) -> str:
        if symbol.find("-") < 0:
            return symbol
        return symbol.split("-")[1]

    def sign(
        self,
        path,
        security_type,
        method_type,
        api_type: Any = "public",
        headers: Optional[Any] = None,
        body: Optional[Any] = None,
        params: Optional[Any] = None,
        config={},
    ):
        host_url = self.apis[self.type]["hostname"]
        destination = self.apis[self.type][security_type][path]["url"]
        version = self.apis[self.type]["version"]
        destination = self.implode_params(destination, params)

        url = host_url + "/" + version + "/" + destination

        if api_type == "private":
            payload = {
                "access_key": self.open_key,
                "nonce": str(uuid.uuid4()),
            }

            if params is not None and len(params) > 0:
                m = hashlib.sha512()
                m.update(urlencode(params).encode())
                payload.update({"query_hash": m.hexdigest(), "query_hash_alg": "SHA512"})

            jwt_token = jwt.encode(payload=payload, key=self.secret_key)
            authorization_token = f"Bearer {jwt_token}"

            if headers is None:
                headers = {}
                headers.update({"content-type": "application/json", "Authorization": authorization_token})

        if method_type.upper() == "POST":
            body = json.dumps(params)
            params = {}

        return {"url": url, "method": method_type, "headers": headers, "body": body, "params": params}

    def get_common_response(self, response):
        if "error" in response:
            return self.create_common_response(
                success="1",
                msg_code=self.safe_string(response["error"], "name"),
                msg=self.safe_string(response["error"], "message"),
                info=response,
            )

        return self.create_common_response(
            success="0",
            msg_code=self.safe_string(response, "name"),
            msg=self.safe_string(response, "message"),
            info=response,
        )

        # if ('status' in response and self.safe_string(response, 'status') == '0000') or ('status' not in response and 'error' not in response):
        #     return self.create_common_response(
        #         success='0',
        #         msg_code=self.safe_string(response, 'status'),
        #         msg=self.safe_string(response, 'message'),
        #         info=response
        #     )
        # else:
        #     return self.create_common_response(
        #         success='1',
        #         msg_code=self.safe_string(response, 'status'),
        #         msg=self.safe_string(response, 'message'),
        #         info=response
        #     )

    def fetch_markets(self, base_market: str = "KRW") -> ksxt.models.KsxtMarketResponse:
        params = {"isDetails": "True"}

        common_header = self.create_common_header(request_params=params)

        response = self.public_get_fetch_markets(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtMarketResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_markets(response=response, base_market=base_market)

        return ksxt.models.KsxtMarketResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_historical_data(
        self, symbol: str, time_frame: str, start: str | None = None, end: str | None = None, base_market: str = "KRW"
    ) -> ksxt.models.KsxtHistoricalDataResponse:
        params = {"market": self.safe_symbol(base_market, symbol), "count": 200}

        # TODO : time_frame 을 어떻게 고정시킬까? 우리는 분봉, 일봉, 주봉, 월봉 만 지원한다고 가정하면?
        if time_frame.endswith("m"):
            # TODO : parse number
            period = 1
            params.update({"unit": period})

            common_header = self.create_common_header(request_params=params)
            response = self.public_get_fetch_security_ohlcv_minute(self.extend(params))

        elif time_frame.endswith("D"):
            common_header = self.create_common_header(request_params=params)
            response = self.public_get_fetch_security_ohlcv_day(self.extend(params))

        elif time_frame.endswith("W"):
            common_header = self.create_common_header(request_params=params)
            response = self.public_get_fetch_security_ohlcv_week(self.extend(params))

        elif time_frame.endswith("M"):
            common_header = self.create_common_header(request_params=params)
            response = self.public_get_fetch_security_ohlcv_month(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtHistoricalDataResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_historical_data(response=response, symbol=symbol, base_market=base_market)

        return ksxt.models.KsxtHistoricalDataResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_ticker(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtTickerResponse:
        params = {"markets": self.safe_symbol(base_market, symbol)}  # 다중 조회 시, 콤마로 구분 ex. 'KRW-BTC, BTC-ETH'

        common_header = self.create_common_header(request_params=params)

        response = self.public_get_fetch_ticker_price(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtTickerResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_ticker(response=response, base_market=base_market)

        return ksxt.models.KsxtTickerResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_tickers(self, symbols: List[str], base_market: str = "KRW") -> ksxt.models.KsxtTickersResponse:
        symbol = ",".join([self.safe_symbol(base_market, symbol) for symbol in symbols])
        params = {"markets": symbol}

        common_header = self.create_common_header(request_params=params)

        response = self.public_get_fetch_tickers_price(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtTickersResponse(header=common_header, response=common_response, info=[])

        # 데이터 파싱
        parsed_info = self.parser.parse_tickers(response, base_market)

        return ksxt.models.KsxtTickersResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_orderbook(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSingleOrderBookResponse:
        params = {"markets": self.safe_symbol(base_market, symbol), "level": 0}

        common_header = self.create_common_header(request_params=params)

        response = self.public_get_fetch_orderbook(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtSingleOrderBookResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_orderbook(response=response, base_market=base_market)

        return ksxt.models.KsxtSingleOrderBookResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_orderbooks(self, symbols: List[str], base_market: str = "KRW") -> ksxt.models.KsxtMultiOrderBookResponse:
        symbol = ",".join([self.safe_symbol(base_market, symbol) for symbol in symbols])
        params = {"markets": symbol}

        common_header = self.create_common_header(request_params=params)

        response = self.public_get_fetch_orderbooks(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtMultiOrderBookResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_orderbooks(response, base_market)

        return ksxt.models.KsxtMultiOrderBookResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_balance(self, acc_num: str, base_market: str = "KRW") -> ksxt.models.KsxtBalanceResponse:
        params = {}

        common_header = self.create_common_header(request_params=params)

        response = self.private_get_fetch_balance(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtBalanceResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_balance(response=response, base_market=base_market)

        return ksxt.models.KsxtBalanceResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_cash(self, acc_num: str, base_market: str = "KRW") -> ksxt.models.KsxtCashResponse:
        params = {}

        common_header = self.create_common_header(request_params=params)

        response = self.private_get_fetch_cash(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtCashResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_cash(response=response, base_market=base_market)

        return ksxt.models.KsxtCashResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_security(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSecurityResponse:
        params = {"market": self.safe_symbol(base_market, symbol)}

        common_header = self.create_common_header(request_params=params)

        response = self.private_get_fetch_security_info(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtSecurityResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_security(response=response, base_market=base_market)

        return ksxt.models.KsxtSecurityResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_trade_fee(self, symbol: Optional[str] = "", base_market: str = "KRW") -> ksxt.models.KsxtTradeFeeResponse:
        params = {"market": self.safe_symbol(base_market, symbol)}

        common_header = self.create_common_header(request_params=params)

        response = self.private_get_fetch_trade_fee(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtTradeFeeResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_trade_fee(response=response, base_market=base_market)
        return ksxt.models.KsxtTradeFeeResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_closed_order_detail(
        self, acc_num: str, order_ids: List[str], base_market: str = "KRW"
    ) -> ksxt.models.KsxtOpenOrderResponse:
        params = {
            "uuids": order_ids,
        }

        common_header = self.create_common_header(request_params=params)

        response = self.private_get_fetch_closed_order_detail(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtOpenOrderResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_closed_order_history(response, base_market)

        return ksxt.models.KsxtOpenOrderResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_open_order_detail(
        self, acc_num: str, order_ids: List[str], base_market: str = "KRW"
    ) -> ksxt.models.KsxtOpenOrderResponse:
        params = {
            "uuids": order_ids,
        }

        common_header = self.create_common_header(request_params=params)

        response = self.private_get_fetch_opened_order_detail(self.extend(params))

        common_response = self.get_common_response(response=response)

        # 실패 시 오류 응답 반환
        if common_response.success != "0":
            return ksxt.models.KsxtOpenOrderResponse(header=common_header, response=common_response, info=None)

        # 데이터 파싱
        parsed_info = self.parser.parse_open_order_history(response, base_market)

        return ksxt.models.KsxtOpenOrderResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_closed_order(
        self,
        acc_num: str,
        symbol: Optional[str] = "",
        start: Optional[str] = None,
        end: Optional[str] = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtClosedOrderResponse:
        params = {"state": "done"}

        if bool(symbol):
            params.update({"market": self.safe_symbol(base_market, symbol)})

        common_header = self.create_common_header(request_params=params)

        response = self.private_get_fetch_closed_order(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtClosedOrderResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_closed_order_history(response=response, base_market=base_market)

        return ksxt.models.KsxtClosedOrderResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_open_order(
        self,
        acc_num: str,
        symbol: str | None = "",
        start: str | None = None,
        end: str | None = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtOpenOrderResponse:
        params = {"state": "wait"}

        if bool(symbol):
            params.update({"market": self.safe_symbol(base_market, symbol)})

        common_header = self.create_common_header(request_params=params)

        response = self.private_get_fetch_opened_order(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtOpenOrderResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_open_order_history(response=response, base_market=base_market)

        return ksxt.models.KsxtOpenOrderResponse(header=common_header, response=common_response, info=parsed_info)

    def fetch_withdrawal_history(
        self, acc_num: str, base_market: str = "KRW"
    ) -> ksxt.models.KsxtWithdrawalHistoryResponse:
        params = {"currency": base_market, "state": "DONE"}

        common_header = self.create_common_header(request_params=params)

        response = self.private_get_fetch_withdrawal_history(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtWithdrawalHistoryResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_withdrawal_history(response=response, base_market=base_market)

        return ksxt.models.KsxtWithdrawalHistoryResponse(
            header=common_header, response=common_response, info=parsed_info
        )

    def fetch_deposit_history(self, acc_num: str, base_market: str = "KRW") -> ksxt.models.KsxtDepositHistoryResponse:
        params = {"currency": base_market, "state": "ACCEPTED"}

        common_header = self.create_common_header(request_params=params)

        response = self.private_get_fetch_deposit_history(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtDepositHistoryResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_deposit_history(response=response, base_market=base_market)

        return ksxt.models.KsxtDepositHistoryResponse(header=common_header, response=common_response, info=parsed_info)

    def create_order(
        self,
        acc_num: str,
        symbol: str,
        ticket_type: Literal["EntryLong", "EntryShort", "ExitLong", "ExitShort"],
        otype: Literal["limit", "market"],
        price: Optional[float] = 0,
        qty: Optional[float] = 0,
        amount: Optional[float] = 0,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtCreateOrderResponse:

        params = {
            "market": self.safe_symbol(base_market, symbol),
        }

        if ticket_type == "EntryLong":
            order_side = "bid"

            if otype == "limit":
                params.update({"side": order_side, "price": str(price), "volume": str(qty), "ord_type": "limit"})

                response = self.private_post_send_order_entry(self.extend(params))
            elif otype == "market":
                params.update(
                    {
                        "side": order_side,
                        "price": amount,
                        "ord_type": "price",
                    }
                )

                response = self.private_post_send_order_entry_market(self.extend(params))
        elif ticket_type == "ExitLong":
            order_side = "ask"
            params = {
                "market": self.safe_symbol(base_market, symbol),
                "side": order_side,
            }

            if otype == "limit":
                params.update({"price": str(price), "volume": str(qty), "ord_type": "limit"})

                response = self.private_post_send_order_exit(self.extend(params))

            elif otype == "market":
                params.update({"ord_type": "market", "volume": str(qty)})

                response = self.private_post_send_order_exit_market(self.extend(params))
        else:
            raise ValueError(
                f"Unsupported ticket_type '{ticket_type}'. {__class__.__name__} supports only 'EntryLong' and 'ExitLong'."
            )

        common_header = self.create_common_header(request_params=params)
        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtCreateOrderResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_create_order(response=response, base_market=base_market)

        return ksxt.models.KsxtCreateOrderResponse(header=common_header, response=common_response, info=parsed_info)

    def cancel_order(
        self,
        acc_num: str,
        order_id: str,
        symbol: str | None = "",
        qty: float = 0,
        *args,
        base_market: str = "KRW",
        **kwargs,
    ) -> ksxt.models.KsxtCancelOrderResponse:
        params = {"uuid": order_id}

        common_header = self.create_common_header(request_params=params)

        response = self.private_delete_send_cancel_order(self.extend(params))

        common_response = self.get_common_response(response=response)
        if common_response.success != "0":
            return ksxt.models.KsxtCancelOrderResponse(header=common_header, response=common_response, info=None)

        parsed_info = self.parser.parse_cancel_order(response=response, base_market=base_market)

        return ksxt.models.KsxtCancelOrderResponse(header=common_header, response=common_response, info=parsed_info)



import asyncio
import aiohttp
import json
import os
import platform
import tomllib
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Literal, Optional

import yarl

from ksxt.async_.base.throttler import Throttler
from ksxt.base.errors import NotSupportedError
from ksxt.base.rest_exchange import RestExchange
from ksxt.config import CONFIG_DIR
import ksxt.models


class AsyncExchange(RestExchange):
    synchronous = False

    def __init__(self, config: Dict = None, filename: str = None):
        super().__init__(config, filename)

        self.asyncio_loop = asyncio.get_event_loop()
        self.session = aiohttp.ClientSession()
        self.throttle = Throttler({}, self.asyncio_loop)

    async def initialize(self):
        if self.asyncio_loop is None:
            self.asyncio_loop = asyncio.get_event_loop()
        if self.session is None:
            self.session = aiohttp.ClientSession()
        if self.throttle is None:
            self.throttle = Throttler({}, self.asyncio_loop)

    async def close(self):
        if self.session:
            await self.session.close()
            self.session = None

    async def __aenter__(self):
        await self.initialize()
        return self

    async def __aexit__(self, *args):
        await self.close()

    def _get_api_from_file(self, filename: str):
        if filename is None:
            tr_config_filename = "tr_dev.json" if self.is_dev else "tr_app.json"
        else:
            tr_config_filename = filename

        config_path = os.path.join(CONFIG_DIR, tr_config_filename)

        if Path(tr_config_filename).suffix == ".json":
            with open(
                config_path,
                encoding="utf-8",
            ) as f:
                c = json.load(f)
                return {"apis": c[self.name]}

        elif Path(tr_config_filename).suffix == ".toml":
            with open(config_path, mode="rb") as f:
                c = tomllib.load(f)
                return c

    async def fetch(self, url, method="GET", headers=None, body=None, params=None):
        request_headers = headers
        request_body = str(body).encode() if body else None
        request_params = params

        session_method = getattr(self.session, method.lower())

        try:
            async with session_method(
                url,
                # yarl.URL(url, encoded=True),
                headers=request_headers,
                data=request_body,
                params=request_params,
                timeout=aiohttp.ClientTimeout(total=int(self.timeout / 1000)),
            ) as response:
                http_response = await response.text(errors="replace")
                json_response = self.parse_json(http_response)
                return json_response
        except asyncio.TimeoutError as e:
            details = f"{self.id} {method} {url}"
            raise TimeoutError(details) from e

    async def fetch2(
        self, path, security_type, params={}, headers: Optional[Any] = None, body: Optional[Any] = None, config={}
    ):
        is_activate = self.apis[self.type][security_type][path]["activate"]
        if not is_activate:
            return {
                "response": {
                    # 성공 실패 여부
                    "success": "-1",
                    # 응답코드
                    "code": "fail",
                    # 응답메세지
                    "message": f"지원하지 않는 함수({path}) 입니다.",
                }
            }

        # if self.enableRateLimit:
        #     cost = self.calculate_rate_limiter_cost(api, method, path, params, config)
        #     self.throttle(cost)

        # self.lastRestRequestTimestamp = self.milliseconds()

        method_type = self.apis[self.type][security_type][path]["method"]
        api_type = self.apis[self.type][security_type][path]["api"]
        request = self.sign(path, security_type, method_type, api_type, headers, body, params, config)
        return await self.fetch(
            request["url"], request["method"], request["headers"], request["body"], request["params"]
        )

    # region base method
    async def fetch_markets(self, market_name: str) -> ksxt.models.KsxtMarketResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_markets.__qualname__}() is not supported yet.")

    async def fetch_security(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSecurityResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_security.__qualname__}() is not supported yet.")

    async def fetch_ticker(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtTickerResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_ticker.__qualname__}() is not supported yet.")

    async def fetch_orderbook(self, symbol: str, base_market: str = "KRW") -> ksxt.models.KsxtSingleOrderBookResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_orderbook.__qualname__}() is not supported yet.")

    async def fetch_historical_data(
        self,
        symbol: str,
        time_frame: str,
        start: Optional[str] = None,
        end: Optional[str] = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtHistoricalDataResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_historical_data.__qualname__}() is not supported yet.")

    async def fetch_is_holiday(self, dt: datetime, base_market: str = "KRW"):
        raise NotSupportedError(f"{self.id} {self.fetch_is_holiday.__qualname__}() is not supported yet.")

    async def fetch_user_info(self, base_market: str = "KRW"):
        raise NotSupportedError(f"{self.id} {self.fetch_user_info.__qualname__}() is not supported yet.")

    async def fetch_balance(self, acc_num: str, base_market: str = "KRW") -> ksxt.models.KsxtBalanceResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_balance.__qualname__}() is not supported yet.")

    async def fetch_cash(self, acc_num: str, base_market: str = "KRW") -> ksxt.models.KsxtCashResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_cash.__qualname__}() is not supported yet.")

    async def fetch_screener_list(self, base_market: str = "KRW"):
        raise NotSupportedError(f"{self.id} {self.fetch_screener_list.__qualname__}() is not supported yet.")

    async def fetch_screener(self, screen_id: str, base_market: str = "KRW"):
        raise NotSupportedError(f"{self.id} {self.fetch_screener.__qualname__}() is not supported yet.")

    async def fetch_deposit_history(
        self, acc_num: str, base_market: str = "KRW"
    ) -> ksxt.models.KsxtDepositHistoryResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_deposit_history.__qualname__}() is not supported yet.")

    async def fetch_withdrawal_history(
        self, acc_num: str, base_market: str = "KRW"
    ) -> ksxt.models.KsxtWithdrawalHistoryResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_withdrawal_history.__qualname__}() is not supported yet.")

    async def create_order(
        self,
        acc_num: str,
        symbol: str,
        ticket_type: Literal["EntryLong", "EntryShort", "ExitLong", "ExitShort"],
        otype: Literal["limit", "market"],
        price: Optional[float] = 0,
        qty: Optional[float] = 0,
        amount: Optional[float] = 0,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtCreateOrderResponse:
        raise NotSupportedError(f"{self.id} {self.create_order.__qualname__}() is not supported yet.")

    async def cancel_order(
        self, acc_num: str, order_id: str, symbol: Optional[str] = "", qty: float = 0, *args, base_market: str = "KRW"
    ) -> ksxt.models.KsxtCancelOrderResponse:
        raise NotSupportedError(f"{self.id} {self.cancel_order.__qualname__}() is not supported yet.")

    async def modify_order(
        self,
        acc_num: str,
        order_id: str,
        price: float,
        qty: float,
        *args,
        symbol: Optional[str] = "",
        base_market: str = "KRW",
    ):
        raise NotSupportedError(f"{self.id} {self.modify_order.__qualname__}() is not supported yet.")

    async def fetch_open_order(
        self,
        acc_num: str,
        symbol: Optional[str] = "",
        start: Optional[str] = None,
        end: Optional[str] = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtOpenOrderResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_open_order.__qualname__}() is not supported yet.")

    async def fetch_closed_order(
        self,
        acc_num: str,
        symbol: Optional[str] = "",
        start: Optional[str] = None,
        end: Optional[str] = None,
        base_market: str = "KRW",
    ) -> ksxt.models.KsxtClosedOrderResponse:
        raise NotSupportedError(f"{self.id} {self.fetch_closed_order.__qualname__}() is not supported yet.")

    async def reserve_order(
        self, acc_num: str, symbol: str, price: float, qty: float, target_date: str, base_market: str = "KRW"
    ):
        raise NotSupportedError(f"{self.id} {self.reserve_order.__qualname__}() is not supported yet.")

    # endregion base method

    async def sleep(self, milliseconds):
        return await asyncio.sleep(milliseconds / 1000)
